/**
	AutonomousActorState:
	AutonomousActorState gathers information along with AutonomousActor
	about the actor's surroundings.  It takes care of finding out about
	an actors processing speed.


	Author: Travis Desell
	Last Modified: 9/19/2002
*/

package io.core.language;

import java.io.*;
import java.net.*;
import salsa.naming.*;
import salsa.naming.exceptions.*;
import salsa.language.*;
import salsa.language.exceptions.*;
import io.core.messaging.TheaterLocator;
import io.core.messaging.IOTheater;

public class AutonomousActorState extends UniversalActorState {
	AutonomousActor self;

	public AutonomousActorState() {
		super();

		self = (AutonomousActor)super.self;
	}

	public AutonomousActorState(UniversalActor _self) {
		super( _self );

		self = (AutonomousActor)super.self;
	}

	public void setSelf(UniversalActor _self) {
		super.setSelf( _self );

		self = (AutonomousActor)super.self;
	}

	public void migrate(String ualString) {
		System.err.println("ActorMigration : " + ualString);

		//set scheduled to false so when it reaches the other theater
		//it may migrate from there
		self.scheduled = false;
		try {
			super.migrate(ualString);
		} catch (MigrationException e) {
			System.err.println("Migration Exception: " + e);
		} catch (UALException e) {
			System.err.println("Improper UAL creation: " + e);
		}

		System.err.println("Migrated: " + ualString);
	}

	public synchronized void process(Message message) {
		long time = System.currentTimeMillis();

		if (! (self).local) { 
			// Actor has moved with its mailbox!
			RunTime.finishedProcessingMessage();
		} else {
			super.process(message);
		}

		((AutonomousActor)self).incrementProcessed(System.currentTimeMillis() - time);
	}
}
