/**

 */
package io.p2p.theater;


import wwc.messaging.*;
import io.p2p.protocol.*;
import java.net.*;
import java.io.*;
import java.util.StringTokenizer;
import java.util.Hashtable;
import java.util.Vector;
import salsa.language.*;
import wwc.util.*;


public class PeerTheater implements Theater {
	// Must be a singleton so UniversalActor can get to it.
	private static PeerTheater instance = new PeerTheater();

	private final static String protocolVersion = RMSProtocol.VERSION;
	private final static String theaterVersion = "PeerTheater/0.1";

	private static Hashtable messageQueue;

	private static Hashtable registry;
	private static ServerSocket passive;
	private static int listenPort = 3030;
	private static int maxConnections = 10;
	private static int informationLevel = 0;
	private static boolean userMode = false;
	private static Vector peerServerList = new Vector();
	private static TheaterLocation theaterLocation;
	private static int maxTimeToLive = 10;
	public  static int maxProcessing = -1;
	public  static int maxMessageSending = -1;

	private static Decider decider;


	/**
		This method creates a passive socket and starts accepting connections to it.
	*/
	private void listen() {
		try {
			passive = new ServerSocket( listenPort );
		} catch (IOException e) {
        		System.err.println( "IOException creating server socket. " + e.getMessage() );

			return;
		}

		listenPort = passive.getLocalPort();
		System.out.println( "Theater listening at: " + theaterLocation.toString() );

		// Loop and accept incoming connections.
		while (true) {
			Socket incoming = null;

		         // Accept the connection.
			try {
				incoming = passive.accept();
			} catch (IOException e) {
				System.err.println( "IOException occured accepting new connection" + e.getMessage() );

				return;
			}

			//Handle the incoming connection
			PeerObjectHandler handler = new PeerObjectHandler(incoming, registry, messageQueue, this, decider, informationLevel, maxTimeToLive);
			handler.start();
		}
	}

	public static void main(String[] args) {
		// Handle arguments...
		getSettings(args);

         	PeerTheater theater = PeerTheater.getInstance();

		try {
			InetAddress IPHere = InetAddress.getLocalHost();

			String StringHere = "rmsp://" + IPHere.getHostAddress() + ":" + listenPort;
			theaterLocation = new TheaterLocation(StringHere);
		} catch (Exception e) {
			System.err.println("Error resolving localhost: " + e);
			System.exit(0);
		}

		//a decider is required
		if ((maxProcessing > -1) && (maxMessageSending > -1)) {
			decider = new Decider(instance, maxProcessing, maxMessageSending, maxConnections);
			decider.start();
		} else {
			PeerTheater.helpMessage("Values for processing (-pr) and bandwidth (-br) must be specified to connect to other peers");
		}


		//	The PeerConnector will connect to the given servers and look for 
		//	The Decider will decide when to migrate actors to other peer theaters, and when to
		//	accept actors from other peer theaters
		if (peerServerList.size() > 0) {
			ServerHandler serverHandler = new ServerHandler(theater, peerServerList, maxConnections);
			serverHandler.start();
		}

		//	The UserListener will listen for user input, and print out the status of the theater
		//		if (userMode) UserListener listener = new UserListener(theater);

         	theater.listen();
   	}

	private static void printSettings() {
		System.out.println("Settings:");
		System.out.println("listenPort = " + listenPort);
		System.out.println("maxTimeToLive = " + maxTimeToLive);
		System.out.println("maxConnections = " + maxConnections);
		System.out.println("informationlevel = " + informationLevel);
		System.out.println("maxProcessing = " + maxProcessing);
		System.out.println("maxMessageSending = " + maxMessageSending);
	}

	private static void getSettings(String[] args) {
		for (int i = 0; i < args.length; i++) {

			if (args[i].equals("-?") || args[i].equals("-h")) 	PeerTheater.helpMessage();
			else if (args[i].equals("-v")) 				PeerTheater.printServerVersion();
			else if (args[i].equals("-user"))			userMode = true;
			else if (args[i].equals("-p") || args[i].equals("-i") || args[i].equals("-c") || args[i].equals("-bw") ||
				 args[i].equals("-pr") || args[i].equals("-ttl") || args[i].equals("-f")) {
				try {
					if (args[i].equals("-p")) 		listenPort = Integer.parseInt(args[i+1]);
					else if (args[i].equals("-i"))		informationLevel = Integer.parseInt(args[i+1]);
					else if (args[i].equals("-c"))		maxConnections = Integer.parseInt(args[i+1]);
					else if (args[i].equals("-pr"))		maxProcessing = Integer.parseInt(args[i+1]);
					else if (args[i].equals("-bw"))		maxMessageSending = Integer.parseInt(args[i+1]);
					else if (args[i].equals("-ttl"))	maxTimeToLive = Integer.parseInt(args[i+1]);
					else if (args[i].equals("-f"))		getSettingsFromFile(args[i+1]);
				} catch (Exception e) {
					if (args[i].equals("-p")) 		PeerTheater.helpMessage("invalid port number : " + args[i+1]);
					else if (args[i].equals("-i"))		PeerTheater.helpMessage("invalid information level : " + args[i+1]);
					else if (args[i].equals("-c"))		PeerTheater.helpMessage("invalid number of max connections : " + args[i+1]);
					else if (args[i].equals("-pr"))		PeerTheater.helpMessage("invalid maximum processing value : " + args[i+1]);
					else if (args[i].equals("-bw"))		PeerTheater.helpMessage("invalid maximum bandwidth : " + args[i+1]);
					else if (args[i].equals("-ttl"))	PeerTheater.helpMessage("invalid maximum time to live : " + args[i+1]);
				}
				i = i + 1;
			}
			else if (args[i].equals("-s")) {
				i++;
				peerServerList = new Vector();
				while ((i < args.length) && ((args[i]).charAt(0) != '-')) {
					peerServerList.add(args[i]);							
					i++;
				}
				i--;
				if (args[i].equals("-s")) PeerTheater.helpMessage("You must specify at least one PeerServer");

			}
			else 							PeerTheater.helpMessage("Invalid Arguments");
		}
	}


	private static void getSettingsFromFile(String filename) {
		FileInputStream file;
		byte input[];

		try {
			file = new FileInputStream(filename);

			input = new byte[file.available()];
			file.read(input);
		} catch (Exception e) {
			System.err.println("File Error: " + e);
			return;
		}

		StringTokenizer st = new StringTokenizer(new String(input), "\n ");
		String args[] = new String[st.countTokens()];

		int i = 0;
		while (st.hasMoreTokens()) {
			args[i] = st.nextToken();
			i++;
		}

		getSettings(args);
	}


	private static void printServerVersion() {
		System.out.println("IO Peer Theater Version: " + theaterVersion);
		System.exit(0);
	}

	private static void helpMessage(String s) {
		System.out.println(s);
		PeerTheater.printHelpMessage();
		System.exit(1);
	}

	private static void helpMessage() {
		PeerTheater.printHelpMessage();
		System.exit(0);
	}

	private static void printHelpMessage() {
		System.out.println("usage :");
		System.out.println("\tjava ...PeerTheater");
		System.out.println("\tjava ...PeerTheater -h");
		System.out.println("\tjava ...PeerTheater -v");
		System.out.println("\tjava ...PeerTheater -p   portNumber");
		System.out.println("\t                    -s   server1 server2 ... serverN");
		System.out.println("\t                    -c   numConnections");
		System.out.println("\t                    -i   informationLevel");
		System.out.println("\t                    -pr  maxProcessingValue(MHZ)");
		System.out.println("\t                    -bw  maxBandwitdh(BPS)");
		System.out.println("\t                    -ttl maxTimeToLive");
		System.out.println("\t                    -f   filename");
		System.out.println("\t                    -user");

		System.out.println("options :");
		System.out.println("\t-h                             : print this message");
		System.out.println("\t-v                             : print version");
		System.out.println("\t-f filename                    : load settings from file");
		System.out.println("\t                                 file is in the same format as command line arguments");
		System.out.println("\t-p portNumber                  : set the listening port to portNumber");
		System.out.println("\t                                 default port number is 3030");
		System.out.println("\t-s server1 server2 ... serverN : gets connections to other");
		System.out.println("\t                                 PeerTheaters, from servers 1 through N");
		System.out.println("\t-ttl maxTimeToLive             : specifies the maximum time to live value of");
		System.out.println("\t                                 packets passing through this theater");
		System.out.println("\t-pr maxProcessValue            : specifies the maximum amount of processing");
		System.out.println("\t                                 ability is availble for this theater, in MHZ");
		System.out.println("\t-bw maxMsgSendingValue         : specifies the maximum amount of bandwidth");
		System.out.println("\t                                 available for this theater, in BPS");
		System.out.println("\t-c numConnections              : specifies the maximum amount of connections");
		System.out.println("\t                                 kept to PeerTheaters, default is 10");
		System.out.println("\t-user                          : has the theater run in user mode, while");
		System.out.println("\t                                 the theater is running, the user may enter");
		System.out.println("\t                                 list and quit commands, to query the state of");
		System.out.println("\t                                 the theater, and to exit");
		System.out.println("\t-i informationLevel            : specifies how informative the theater is:");
		System.out.println("\t                                 0 - nothing is displayed (default)");
		System.out.println("\t                                 1 - migration of actors displayed");
		System.out.println("\t                                 2 - migration of actors and p2p broadcasts displayed");
		System.out.println("\t                                 3 - migration of actors, p2p broadcasts and messages displayed");
	}

	/**
		The following functions are used by autonomous and universal actors to access
		various information found in the theater
	*/
	public static PeerTheater getInstance() {
		return instance;
	}

	public TheaterLocation getLocation() {
		return theaterLocation;
	}

	public void removeAutonomousActor(AutonomousActor a) {
		if (decider != null) {
			decider.removeActor(a);
		}
	}

	public Hashtable getRegistry() {
		return registry;
	}

	private PeerTheater() {
		registry = new Hashtable();
		messageQueue = new Hashtable();
	}

   	public int getPort() {
		return listenPort;
	}

   	public UniversalActor getActorUAL( String relativeUAL ) {
		//This returns a UniversalActor, given a particular UAL
		return (UniversalActor)registry.get(relativeUAL);
	}
}
