/** SALSA/World Wide Computer Project
 *  ObjectHandler recieves incoming objects
 *  and determines if they are actors or messages.
 *  upon recieving actors, it will start them up.
 *  messages get sent to the appropriate actor.
 *
 *  author Travis Desell
 */
package salsa.messaging;

import java.net.*;
import java.io.*;
import java.util.Hashtable;
import salsa.language.*;
import salsa.messaging.services.*;

public class ObjectHandler extends Thread {
	// Must be a singleton so UniversalActor can get to it.
	private Socket incoming;
	private Hashtable registry;
	private Theater theater;

	public ObjectHandler(Socket _incoming, Hashtable _registry, Theater _theater) {
		incoming = _incoming;
		registry = _registry;
		theater = _theater;

		setName( "Object Handler Thread" );
	}

   /**
      This method creates a passive socket and starts accepting connections to it.
      it contains some of the functionality once found in ObjectInputGate.
   */
	public void run() {
		// Read in the object.
		Object received = null;
		try {
			ObjectInputStream stream = new ObjectInputStream( incoming.getInputStream() );
			received = stream.readObject();
		} catch (ClassNotFoundException e) {
			System.err.println( "Could not load class for " + received.getClass().getName() );
			System.err.println( "  Is it in the theater CLASSPATH" );
			System.err.println( e.getMessage() );

			return;
		} catch (IOException e) {
			System.err.println( "IOException occured reading object " + e.getMessage() );

			return;
		}

		// Close the connection.
		try {
			incoming.close();
		} catch (IOException e) {
			System.err.println( "IOException occured closing connection" + e.getMessage() );

			return;
		}

		// Process the object.
		process( received );
	}

	/**
		This method contains control logic for deciding what to do
      		with a read object.
   	*/
   	private void process(Object o) {
      		// If it's a message to be passed on...
		if (o instanceof MigrationWrapper) {
        	 	setPriority(MAX_PRIORITY);

			UniversalActor incomingActor = ((MigrationWrapper) o).getActor();

			// Update any references if an actor reference already exists.
			incomingActor.resolveActorState( theater );

        	 	// Add the actor to the registry.
        	 	String relativeUAL = incomingActor.getUAL().getIdentifier();
        	 	registry.put(relativeUAL, incomingActor);

			RunTime.receivedUniversalActor();
		}

      		if (o instanceof Message) {  // just serializing data
			setPriority(MIN_PRIORITY);
			yield();

	        	Message m = (Message) o;

			// Resolve the sender and target.
			m.resolveActors( theater );
				
			// Send the message.
       			m.getTarget().send(m);
      		}
   	}
}
