
package io.p2p.theater;

import java.util.Vector;
import io.p2p.protocol.*;
import salsa.language.*;

public class Decider extends Thread {
	private PeerTheater theater;
	private int maxProcessing;
	private int maxMessageSending;
	private int maxTheaters;
	private boolean dontAccept = false;
	private boolean haveRequested = false;
	private Vector peerTheaters;
	private Vector actors;

	private Vector incomingTheaterQueue;
	private int maxTheaterQueueSize = 50;

	private long currentMessageId = 0;

	private long peerTheaterTimer;
	private int lastActors = 0;

	public int totalProcessed = 0;
	public int totalSent = 0;
	public int iterations = 0;

	public int getMaxProcessing() { return maxProcessing; }
	public int getMaxMessageSending() { return maxMessageSending; }

	public Decider(PeerTheater _theater, int _maxProcessing, int _maxMessageSending, int _maxTheaters) {
		theater = _theater;
		maxProcessing = _maxProcessing;
		maxMessageSending = _maxMessageSending;
		peerTheaters = new Vector();
		actors = new Vector();
		maxTheaters = _maxTheaters;

		incomingTheaterQueue = new Vector();

		peerTheaterTimer = System.currentTimeMillis();
	}

	public Vector getIncomingTheaterQueue() {
		return incomingTheaterQueue;
	}

	public int getMaxTheaterQueueSize() {
		return maxTheaterQueueSize;
	}

	public Vector getTheaterList() {
		return peerTheaters;
	}

	public synchronized void addActor(AutonomousActor a) {
		actors.add(a);
		reset();
	}

	public synchronized void removeActor(AutonomousActor a) {
		actors.remove(a);
		reset();
	}


	public double getAvailableTime() {
		double timeUsed = 0;

		for (int i = 0; i < actors.size(); i++) {
			AutonomousActor a = (AutonomousActor)actors.get(i);			

			timeUsed += a.getProcessingTime() + a.getSendingTime();
		}

		return (System.currentTimeMillis() - peerTheaterTimer) - timeUsed;
	}

	//hasTheater is used to check and see if the theaterList contains a certain theater
	public boolean hasTheater(TheaterLocation theaterLocation) {
		for (int i = 0; i < peerTheaters.size(); i++) {
			if ( ((TheaterLocation)peerTheaters.get(i)).equals(theaterLocation) ) {
				return true;
			}
		}
		return false;
	}


	//add theater is used when the theater recieves a Pong packet
	public void addTheater(TheaterLocation theaterLocation, long pingTime) {
		theaterLocation.pingTime = pingTime;

		long slowestPing = -1;
		int slowestTheater = -1;

		for (int i = 0; i < peerTheaters.size(); i++) {
			long currentPing = ((TheaterLocation)peerTheaters.get(i)).pingTime;

			if (currentPing == -1) {
				peerTheaters.remove(i);
				peerTheaters.add(theaterLocation);
				return;
			}

			if (currentPing > slowestPing) {
				slowestPing = currentPing;
				slowestTheater = i;
			}
		}

		if (slowestPing > -1) {
			peerTheaters.remove(slowestTheater);
			peerTheaters.add(theaterLocation);
		}

	}

	//addTheater without any arguments except the theater is used when the theater is first
	//populating its list of theaters
	public void addTheater(TheaterLocation theaterLocation) {
		if (peerTheaters.size() < maxTheaters) peerTheaters.add(theaterLocation);
	}

	//migrationAccepted is used when the theater recieves a
	//MigrationAccept packet
	public boolean migrationAccepted(double foreignProcessing, double foreignMessageSending, double foreignTime, TheaterLocation theaterLocation) {
		//get the actor with the closest fit to the processing and messageSending sent

		AutonomousActor currentActor;
		AutonomousActor bestActor = null;

		double bestGain = -1;
		double currentGain = -1;

		System.out.println("Running the decision function, foreignProcessing: " + foreignProcessing + 
				   ", foreignMessageSending: " + foreignMessageSending);

		//get the actor that will benefit the most by migrating
		for (int i = 0; i < actors.size(); i++) {
			currentActor = (AutonomousActor)actors.get(i);

			double foreignMessagesProcessed = 0;

			if (currentActor.getMessagesProcessed() > 0) {
				//Find the average cycles per message processed
				double cycles = currentActor.getProcessingTime() * maxProcessing / currentActor.getMessagesProcessed();

				//Find the time it will take to process single messages on the foreign theater
				double foreignProcessTime = cycles / foreignProcessing;

				double singleFullProcessTime = 0;

				if (currentActor.getNonLocalMessagesSent() > 0) {
					//Find the average bits per message sent to a foreign theater
					double bits = currentActor.getSendingTime() * maxMessageSending / currentActor.getNonLocalMessagesSent();

					//Find the time it will take to send single messages on the foreign theater
					double foreignSendingTime = bits / foreignMessageSending;
	
					//To fully process one message (including sending what it sends) on foreign theater
					double sendingRatio = currentActor.messagesNotSentTo(theaterLocation) / currentActor.getMessagesProcessed();
					singleFullProcessTime = (foreignProcessTime + (sendingRatio * foreignSendingTime));
				} else {
					singleFullProcessTime = foreignProcessTime;
				}

				foreignMessagesProcessed = foreignTime / singleFullProcessTime;
			} else {
				if (currentActor.actorState != null) {
					if (currentActor.mailboxSize() > 0) {
						foreignMessagesProcessed = 1;
					} else foreignMessagesProcessed = 0;
				} else {
					foreignMessagesProcessed = 0;
				}
			}

			currentGain = foreignMessagesProcessed - currentActor.getMessagesProcessed();

			System.out.println("CurrentGain = " + currentGain);

			if ((currentGain > 0)&&(currentGain > bestGain)) {
				if (currentActor.actorState != null) {
					bestActor = currentActor;
					bestGain = currentGain;
				}
			}
		}

		if (currentGain > 0) {
			//have our best candidate actor migrate to the theater, and send an accept message
			Actor[] _targets = { bestActor };
			String[] _methodNames = { "migrate" };
			Object[][] _arguments = { {theaterLocation} };

			bestActor.prioritySend( new Message( bestActor, _targets, _methodNames, _arguments ) );

			Accept accept = new Accept();
			accept.send(theaterLocation);

			return true;
		} else {
			return false;
		}
	}

	//willAccept and responding is used when the theater recieves a MigrationRequest packet
	public boolean willAccept(double processingValue, double messageSendingValue) {
		return ( (getAvailableTime() > 0) && (!dontAccept));
	}

	public void responding() {
		dontAccept = true;
	}

	//Accept and Reject both use respondingFinished to let the theater know it may
	//resume sending out migrationAccept packets
	public void respondingFinished() {
		dontAccept = false;
	}

	//this resets all the theaters and actors timers
	public void reset() {
		peerTheaterTimer = System.currentTimeMillis();

		for (int i = 0; i < actors.size(); i++) {
			AutonomousActor a = (AutonomousActor)actors.get(i);

			totalSent += a.getMessagesSent();
			totalProcessed += a.getMessagesProcessed();			

			a.reset();
		}
	}

	public boolean isFull() {
		for (int i = 0; i < actors.size(); i++) {
			AutonomousActor a = (AutonomousActor)actors.get(i);

			if (!a.satisfied()) return true;
		}
		return false;
	}

	public AutonomousActor getLeastSatisfied() {
		double satisfaction = 0;
		AutonomousActor selectedActor = null;

		for (int i = 0; i < actors.size(); i++) {
			AutonomousActor a = (AutonomousActor)actors.get(i);

			if (!a.satisfied()) {
				double currentSatisfaction = 0;
				if (a.getMessagesRecieved() != 0) {
					currentSatisfaction = a.getMessagesProcessed()/a.getMessagesRecieved();
				}

				if (currentSatisfaction > satisfaction) {
					satisfaction = currentSatisfaction;
					selectedActor = a;
				}
			}
		}
		return selectedActor;
	}

	public void printInformation() {
		int sent = 0;
		int processed = 0;

		System.out.println("\nTheater Information, after " + iterations + " iterations:");

		for (int i = 0; i < actors.size(); i++) {
			AutonomousActor a = (AutonomousActor)actors.get(i);

//			a.printStatistics();

			sent += a.getMessagesSent();
			processed += a.getMessagesProcessed();
		}

//		System.out.println("Messages Sent since last reset: " + sent);
//		System.out.println("Messages Processed since last reset: " + processed + "\n");

		System.out.println("Total Sent: " + (totalSent + sent));
		System.out.println("Total Processed: " + (totalProcessed + processed));

		iterations++;
	}


	public void run() {

		try {
			sleep(30000);	//sleep 30 seconds before sending an available message to the gathered theaters
		} catch (InterruptedException e) {
			System.err.println("Sleep Interrupted: " + e);
		}

		Available available = new Available(maxProcessing, 
						    maxMessageSending, 
						    getAvailableTime(),
						    new TheaterLocation(theater.getLocation(), currentMessageId), 5);
		available.broadcast(peerTheaters);
		currentMessageId++;

		while (true) {
			try {
				sleep(10000);
			} catch (InterruptedException e) {
				System.err.println("Sleep Interrupted: " + e);
			}

			if (isFull()) {
//				System.out.println("Theater is full");

				if (!haveRequested) {
					AutonomousActor leastSatisfied = getLeastSatisfied();

					if (leastSatisfied != null) {
						MigrationRequest migrationRequest = new MigrationRequest(leastSatisfied.getProcessingTime()*maxProcessing,
													 leastSatisfied.getSendingTime()*maxMessageSending,
													 new TheaterLocation(theater.getLocation(), currentMessageId),
													 2);
						migrationRequest.broadcast(peerTheaters); 
						currentMessageId++;
						haveRequested = true;
					}
				}
			} else {
				haveRequested = false;
//				System.out.println("Theater is not full");
			}

			printInformation();
		}

	}
}
